#include <Rcpp.h>
using namespace Rcpp;

// Function to run the temperature index swe model for one elevation band
// Philip Kraaijenbrink
// 2021-04-23
// v1.0.0

// Additional components:
//   - Snow refreezing component
//   - Advection of rain-on-snow 
//   - Albedo decay
//   - Snow depth estimation


// Function arguments:
// tas        vector of temperatures for each time step
// prcp       vector of precipitation for each time step
// monind     vector of indices for the current month for each time step
// lapserate  vector of temperature lapse rate for every month (n=12)
// zdiff      numeric of elevation difference of the elevation band
//            with the mean elevation of the tas forcing cell (i.e. ERA5)
// bandbias   numeric of T bias of this zband with respect to T forcing
// dff_sm     numeric of degree day factor for snow
// toffset    vector of T offset values for each month
// pscale     vector of precipitation delta factors for each month
// optrun     boolean flag to output optimization data only, i.e. snowdays
// tlimlo     numeric of limit below which all precip is solid
// tlimhi     numeric of limit above which all precip is liquid
// ddf_rf     numeric of degree day factor for snow melt refreezing
// steplen    numeric of length of timestep in days
// calcdepth  boolean flag whether to perform snow depth calculations

// [[Rcpp::export]]
List sweCPP(NumericVector tas, NumericVector prcp,
            NumericVector monind, NumericVector lapserate,
            double zdiff, double bandbias,
            double ddf_sm_lo, double ddf_sm_hi,
            NumericVector toffset, NumericVector pscale, bool optrun,
            double tlimlo=0.0, double tlimhi=2.0,
            double snowsc=0.1, double ddf_rf=1.0,
            double steplen=1.0, bool calcdepth=0){

  // define variables
  int    n                = tas.size();

  double tband            = 0;
  double precip           = 0;
  double trange           = tlimhi-tlimlo;

  double potmelt          = 0;
  double actmelt          = 0;
  double refreeze         = 0;
  double snowdays         = 0;

  double snowfall         = 0;
  double rainfall         = 0;
  double rainonsnow       = 0;
  double rosmelt          = 0;

  double snowsum          = 0;
  double rainsum          = 0;
  double rossum           = 0;

  // albedo related vars
  double dd_since_snow    = 0;
  double ddf_sm           = 0;
  double albedo           = 0;
  double a_lo             = 0.40;
  double a_hi             = 0.85;

  // snow depth vars
  NumericVector snowdepth (n);
  double density_fresh    = 0;
  double snowdensity      = 0;
  double viscosity        = 0;
  double overburden       = 0;
  double compaction       = 0;
  // params (from saloranta, 2016)
  double max_compaction   = 0.5;
  double density_max      = 0.6;
  double dens_fresh_min   = 0.05;
  double n0               = 7.622;
  double c5               = 0.1;
  double c6               = 24.3;

  double tmean            = 0;

  NumericVector snowpack (n);
  NumericVector snowstore (n);
  NumericVector watstore (n);
  NumericVector icestore (n);

  NumericVector qsnow (n);


  // loop over timesteps, start at step 2
  for (int i = 1; i < n; ++i) {

    // get meteo conditions ============================================

    // get air temperature for elevation band under consideration
    tband = tas[i] + bandbias + zdiff * lapserate[monind[i]-1] + toffset[monind[i]-1];

    // determine snow and rain from P and T, linear scale between tlow and thigh
    precip = prcp[i] * pscale[monind[i]-1];
    if (tband < tlimlo){
      snowfall      = precip;
    }else if (tband > tlimhi){
      snowfall = 0.0;
    }else{
      snowfall = (1.0 - tband / trange) * precip;
    }
    rainfall   = precip - snowfall;
    rainonsnow = (snowstore[i-1] > 0) ? rainfall : 0;

    // update precip trackers
    snowsum = snowsum + snowfall;
    rainsum = rainsum + rainfall;
    rossum  = rossum + rainonsnow;

    // update snow and ice storage ==============================================

    // when its freezing:
    if (tband < 0.0){

      // get actual refreezing
      refreeze = std::min(-tband * ddf_rf * steplen,  watstore[i-1]);

      // update storage
      snowstore[i] = snowstore[i-1] + snowfall;
      icestore[i]  = icestore[i-1]  + refreeze;
      watstore[i]  = watstore[i-1]  - refreeze;

    // when its thawing:
    }else{

      // determine albedo decay effects using degree days (following Brock, 2000)
      dd_since_snow = snowfall > 0 ? 0 : dd_since_snow + tband*steplen;
      albedo        = 0.713 - 0.112 * log10(dd_since_snow);
      albedo        = albedo > a_hi ? a_hi : albedo;
      albedo        = albedo < a_lo ? a_lo : albedo;

      // convert decayed albedo to DDF (inversely scale to min max ddf range)
      ddf_sm = (((1-albedo)-(1-a_hi)) * (ddf_sm_hi-ddf_sm_lo)) / ((1-a_lo)-(1-a_hi)) + ddf_sm_lo;

      // get melt
      rosmelt = rainonsnow * tband * 4.186 / 333.55;
      potmelt = tband * ddf_sm * steplen + rosmelt;
	    actmelt = std::min(snowpack[i-1], potmelt);

      // first melt the top snow layer
      snowstore[i] = snowstore[i-1] - actmelt;

      // melt the refreezed ice layer if all top snow is gone
      icestore[i] = icestore[i-1] + std::min(0.0, snowstore[i]);

      // snow and ice storage are capped at zero
      snowstore[i] = std::max(0.0, snowstore[i]);
      icestore[i]  = std::max(0.0, icestore[i]);

	    // add any snow that might fall when its thawing (i.e. between tlow-thigh)
	    snowstore[i] = snowstore[i] + snowfall;

      // set water storage capacity and fill water storage with melt water
      watstore[i] =  std::min(snowsc*snowstore[i], watstore[i-1]+actmelt);

      // determine snow melt runoff, i.e. actual melt water and/or relase of watstore undercapacity
      qsnow[i] = actmelt - (watstore[i] - watstore[i-1]);
    }

    // get final derivatives  ============================================

    // calculate total snow pack, i.e. SWE
    snowpack[i] = snowstore[i] + icestore[i];

    // accumulate total snowdays
    if (snowpack[i]>0){
      snowdays += 1;
    }

    // get mean T
    tmean = (tmean + tband) / 2;



    // snow depth calculations ============================================
    // (following seNorge, saloranta 2012. Uses updated params from saloranta 2019)
    if (calcdepth){

      // if there is snowmelt, reduce snowdepth using snowpack ratio
      if (snowpack[i-1] > 0.0){  // this should be pack-pack[i-1]????
        snowdepth[i] = snowdepth[i-1] * std::min(std::max(snowpack[i]-snowfall, 0.0) / snowpack[i-1], 1.0);
      }else{
        snowdepth[i] = snowdepth[i-1];
      }

      // add fresh snow to depth using fresh density
      density_fresh = dens_fresh_min + std::pow(std::max(0.0, tband*1.8+32)/100, 2);
      snowdepth[i]  = snowdepth[i] + snowfall/density_fresh;

      // perform gradual compaction
      if (snowdepth[i]>0){
        snowdensity  = snowpack[i]/snowdepth[i];
        overburden   = 0.5 * 9.81 * snowpack[i];
        viscosity    = 1/(1+60*watstore[i]/snowdepth[i]) * snowdensity/0.250 * n0*1e6 * std::exp(-c5*std::min(0.5*tband,0.0) + c6*snowdensity);
        compaction   = overburden/viscosity * snowpack[i] * steplen*86400;
        if (max_compaction*snowdepth[i] < compaction){
          compaction = max_compaction*snowdepth[i];
        }
        snowdepth[i] = snowdepth[i]-compaction;
      }
      if ((snowpack[i]/snowdepth[i])>0.6){
        snowdepth[i] = snowpack[i]/density_max;
      }
    }

  }


  // control output =================================================

  // scale number of snow days with the time step length
  snowdays = snowdays * steplen;

  // define output based on function argument
  if (optrun){
    return List::create(Named("snowdays")  = snowdays);
  } else if (calcdepth){
    return List::create(Named("snowpack")  = snowpack,
                        Named("discharge") = qsnow,
                        Named("snowdepth") = snowdepth,
                        Named("snow_sum")  = snowsum,
                        Named("rain_sum")  = rainsum,
                        Named("ros_sum")   = rossum,
                        Named("tas_mean")  = tmean);
  } else {
    return List::create(Named("snowpack")  = snowpack,
                        Named("discharge") = qsnow,
                        Named("snow_sum")  = snowsum,
                        Named("rain_sum")  = rainsum,
                        Named("ros_sum")   = rossum,
                        Named("tas_mean")  = tmean);
  }
}
